// Lobster: tools for biological sequence analysis
// (C) Copyright 2002-2003 Robert C. Edgar
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANT-
// ABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
// Public License for more details.

#include "lobster.h"
#include "HMM.h"
#include "HMMPath.h"
#include "Seq.h"

#define	VERBOSE	0

// Macros to simulate 2D matrices
#define DPM(NodeIndex, PrefixLength)	DPM_[(NodeIndex)*(uPrefixCount) + (PrefixLength)]
#define DPD(NodeIndex, PrefixLength)	DPD_[(NodeIndex)*(uPrefixCount) + (PrefixLength)]
#define DPI(NodeIndex, PrefixLength)	DPI_[(NodeIndex)*(uPrefixCount) + (PrefixLength)]

SCORE HMM::ForwardSeq(const Seq &s, SCORE **ptrptrDPM)
	{
	const unsigned uSeqLength = s.Length();
	const unsigned uPrefixCount = uSeqLength + 1;
	const unsigned uNodeCount = GetNodeCount();
	assert(uSeqLength > 0 && uNodeCount > 0);
	*ptrptrDPM = 0;

// Allocate DP matrices
	size_t LM = uPrefixCount*uNodeCount;
	SCORE *DPM_ = new SCORE[LM];
	SCORE *DPD_ = new SCORE[LM];
	SCORE *DPI_ = new SCORE[LM];

#if	_DEBUG
	memset(DPM_, 0xcd, LM*sizeof(SCORE));
	memset(DPD_, 0xcd, LM*sizeof(SCORE));
	memset(DPI_, 0xcd, LM*sizeof(SCORE));
#endif

	DPM(0, 0) = MINUS_INFINITY;
	DPI(0, 0) = MINUS_INFINITY;
	DPD(0, 0) = m_scoreFirstD;

	for (unsigned uNodeIndex = 1; uNodeIndex < uNodeCount; ++uNodeIndex)
		{
		DPM(uNodeIndex, 0) = MINUS_INFINITY;
		DPD(uNodeIndex, 0) = Add2(DPD(uNodeIndex-1, 0), GetNode(uNodeIndex-1).m_scoreDD);
		DPI(uNodeIndex, 0) = MINUS_INFINITY;
		}

	const HMMNode &Node0 = GetNode(0);
	for (unsigned uPrefixLength = 1; uPrefixLength < uPrefixCount; ++uPrefixLength)
		{
		assert(uPrefixLength > 0);
		const unsigned uLetter = s.GetLetter(uPrefixLength-1);

	// Node 0 is special case.
	// Global/global
		if (1 == uPrefixLength)
			DPM(0, uPrefixLength) = m_scoreFirstM + Node0.m_scoreMatchEmit[uLetter];
		else
			DPM(0, uPrefixLength) = MINUS_INFINITY;
		DPD(0, uPrefixLength) = MINUS_INFINITY;
		SCORE scoreEmit = Node0.m_scoreInsertEmit[uLetter];
		SCORE scoreMI = Add3(DPM(0, uPrefixLength-1), Node0.m_scoreMI, scoreEmit);
		SCORE scoreDI = Add3(DPD(0, uPrefixLength-1), Node0.m_scoreDI, scoreEmit);
		SCORE scoreII = Add3(DPI(0, uPrefixLength-1), Node0.m_scoreII, scoreEmit);
		DPI(0, uPrefixLength) = SumLog(scoreMI, scoreDI, scoreII);

		for (unsigned uNodeIndex = 1; uNodeIndex < uNodeCount; ++uNodeIndex)
			{
			assert(uNodeIndex > 0);
			assert(uPrefixLength > 0);

			const HMMNode &Node = GetNode(uNodeIndex);
			const HMMNode &PrevNode = GetNode(uNodeIndex-1);

		// Transitions into M
			{
			SCORE scoreEmit = Node.m_scoreMatchEmit[uLetter];
			SCORE scoreMM = Add3(DPM(uNodeIndex-1, uPrefixLength-1), PrevNode.m_scoreMM,
			  scoreEmit);
			SCORE scoreDM = Add3(DPD(uNodeIndex-1, uPrefixLength-1), PrevNode.m_scoreDM,
			  scoreEmit);
			SCORE scoreIM = Add3(DPI(uNodeIndex-1, uPrefixLength-1), PrevNode.m_scoreIM,
			  scoreEmit);
			DPM(uNodeIndex, uPrefixLength) = SumLog(scoreMM, scoreDM, scoreIM);
			}

		// Transitions into D
			{
			SCORE scoreMD = Add2(DPM(uNodeIndex-1, uPrefixLength), PrevNode.m_scoreMD);
			SCORE scoreDD = Add2(DPD(uNodeIndex-1, uPrefixLength), PrevNode.m_scoreDD);
			SCORE scoreID = Add2(DPI(uNodeIndex-1, uPrefixLength), PrevNode.m_scoreID);
			DPD(uNodeIndex, uPrefixLength) = SumLog(scoreMD, scoreDD, scoreID);
			}

		// Transitions into I
			{
			SCORE scoreEmit = Node.m_scoreInsertEmit[uLetter];
			SCORE scoreMI = Add3(DPM(uNodeIndex, uPrefixLength-1), Node.m_scoreMI,
			  scoreEmit);
			SCORE scoreDI = Add3(DPD(uNodeIndex, uPrefixLength-1), Node.m_scoreDI,
			  scoreEmit);
			SCORE scoreII = Add3(DPI(uNodeIndex, uPrefixLength-1), Node.m_scoreII,
			  scoreEmit);
			DPI(uNodeIndex, uPrefixLength) = SumLog(scoreMI, scoreDI, scoreII);
			}
			}
		}

	const unsigned uLastNodeIndex = uNodeCount - 1;
	SCORE scoreM = DPM(uLastNodeIndex, uSeqLength);
	SCORE scoreD = DPD(uLastNodeIndex, uSeqLength);
	SCORE scoreFwd = SumLog(scoreM, scoreD);

#if VERBOSE
	ListDPProb(uNodeCount, uSeqLength, DPM_, DPD_, DPI_);
#endif	// VERBOSE

//	delete[] DPM_;
	*ptrptrDPM = DPM_;
	delete[] DPD_;
	delete[] DPI_;

	return scoreFwd;
	}

void ListDP(unsigned uNodeCount, unsigned uSeqLength, const SCORE *DPM_,
  const SCORE *DPD_, const SCORE *DPI_)
	{
	const unsigned uPrefixCount = uSeqLength + 1;
	List("Prefix   ");
	for (unsigned uPrefixLength = 0; uPrefixLength < uPrefixCount; ++uPrefixLength)
		List(" %8u", uPrefixLength);
	List("\n");
	List("M[]\n");
	for (unsigned uNodeIndex = 0; uNodeIndex < uNodeCount; ++uNodeIndex)
		{
		List("Node %u:  ", uNodeIndex);
		for (unsigned uPrefixLength = 0; uPrefixLength < uPrefixCount; ++uPrefixLength)
			List(" %s", ScoreToStr(DPM(uNodeIndex, uPrefixLength)));
		List("\n");
		}
	List("D[]\n");
	for (unsigned uNodeIndex = 0; uNodeIndex < uNodeCount; ++uNodeIndex)
		{
		List("Node %u:  ", uNodeIndex);
		for (unsigned uPrefixLength = 0; uPrefixLength < uPrefixCount; ++uPrefixLength)
			List(" %s", ScoreToStr(DPD(uNodeIndex, uPrefixLength)));
		List("\n");
		}
	List("I[]\n");
	for (unsigned uNodeIndex = 0; uNodeIndex < uNodeCount; ++uNodeIndex)
		{
		List("Node %u:  ", uNodeIndex);
		for (unsigned uPrefixLength = 0; uPrefixLength < uPrefixCount; ++uPrefixLength)
			List(" %s", ScoreToStr(DPI(uNodeIndex, uPrefixLength)));
		List("\n");
		}
	}

static void PutScore(SCORE Score)
	{
	if (MINUS_INFINITY == Score)
		List("        *");
	else
		List("%9.5f", ScoreToProb(Score));
	}

void ListDPProb(unsigned uNodeCount, unsigned uSeqLength, const SCORE *DPM_,
  const SCORE *DPD_, const SCORE *DPI_)
	{
	const unsigned uPrefixCount = uSeqLength + 1;
	List("Prefix   ");
	for (unsigned uPrefixLength = 0; uPrefixLength < uPrefixCount; ++uPrefixLength)
		List(" %8u", uPrefixLength);
	List("\n");
	List("M[]\n");
	for (unsigned uNodeIndex = 0; uNodeIndex < uNodeCount; ++uNodeIndex)
		{
		List("Node %u:  ", uNodeIndex);
		for (unsigned uPrefixLength = 0; uPrefixLength < uPrefixCount; ++uPrefixLength)
			PutScore(DPM(uNodeIndex, uPrefixLength));
		List("\n");
		}
	List("D[]\n");
	for (unsigned uNodeIndex = 0; uNodeIndex < uNodeCount; ++uNodeIndex)
		{
		List("Node %u:  ", uNodeIndex);
		for (unsigned uPrefixLength = 0; uPrefixLength < uPrefixCount; ++uPrefixLength)
			PutScore(DPD(uNodeIndex, uPrefixLength));
		List("\n");
		}
	List("I[]\n");
	for (unsigned uNodeIndex = 0; uNodeIndex < uNodeCount; ++uNodeIndex)
		{
		List("Node %u:  ", uNodeIndex);
		for (unsigned uPrefixLength = 0; uPrefixLength < uPrefixCount; ++uPrefixLength)
			PutScore(DPI(uNodeIndex, uPrefixLength));
		List("\n");
		}
	}
